/*
 * REMARK: If the configuration of SATFO is such that output is
 *         written to the screen, the device 'stderr' is closed on
 *         termination of the SA loop.
 *         Consequently, no other ouput will appear on the screen which
 *         is produces by code called after SATFO.
 *         This will be repaired in the future.
 */

/********************************************************************************
 *                                                                              *
 *  Simulated Annealing Toolbox For Optimization (SATFO)                        *
 *  is written in ANSI standard C and it comprises a part of the                *
 *  Ph.D. research conducted by Antoine H.C. van Kampen during                  *
 *  1992-1996                                                                   * 
 *                                                                              *
 *  Written by A.H.C. van Kampen                                                *
 *                                                                              *
 *                                                                              *
 *                                                                              *
 *                                                                              *
 *                                                                              *
 *  SATFO is free software: you can redistribute it and/or modify               *
 *  it under the terms of the GNU Lesser General Public License as published by *
 *  the Free Software Foundation, either version 3 of the License, or           *
 *  (at your option) any later version.                                         *
 *                                                                              *
 *  SATFO is distributed in the hope that it will be useful,                    *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of              *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               * 
 *  GNU Lesser General Public License for more details.                         *
 *                                                                              *
 *  You should have received a copy of the GNU Lesser General Public License    *
 *  along with SATFO.  If not, see <http://www.gnu.org/licenses/>.              *  
 *                                                                              *
 *                                                                              *
 *                                                                              * 
 *                                                                              *
 *                                                                              *
 *                                                                              *
 *******************************************************************************/

/************************************************************************
 *                                                                      *
 * This is Version 4.0 of SATFO.                                        *
 *                                                                      *
 * Changes from version 3.5:                                            *
 *        - The array centers is removed                                *
 *        - Output file for number of function evaluations is included. *
 *        - Best parameter set is written to file every Markove chain   *
 *        - Acceptance ratio and temperature are written with           *
 *          increase precision to file and screen                       *
 ************************************************************************/

/************************************************************************
 *                                                                      *
 * Calling: SATFO is used by writing your own cost function (and        *
 *          generating function) in a seperate file and then calling    *
 *          the main function SA(argc, argv). So your file would look   *
 *          this:                                                       *
 *              void main (int argc, char *argv[])                      *
 *		{                                                       *
 *		 SA(argc, argv);                                        *
 *              }                                                       *
 *                                                                      *
 *              EvaluteCostFunction(int NumberOfParameters,             *
 *                                  double *ParamValues)                *
 *              {                                                       *
 *                your own cost function                                *
 *              }                                                       *
 *                                                                      *
 *              GenerateNewParametersUser(double *ParametersNew,        *
 *                                        double *ParametersOld,        *
 *                                        double *UpperBounds,          *
 *	                                  double *LowerBounds,          *
 *                                        int     NumberOfParameters,   *
 *                                        double  ControlParameter)     *
 *              {                                                       *
 *                your own generating function                          *
 *                You can also use one of the standard functions        *
 *              }                                                       *
 *                                                                      *
 ************************************************************************/

/************************************************************************
 *                                                                      *
 * Compilation: if any modifications are made to SA.c then first        *
 *              compile this module with gcc -ansi -c SA.c.             *
 *              If you made any changes to your user module (the        *
 *              calling program) the recompile with:                    *
 *              gcc -ansi -o simaneal <user module> SA.o -lm            *
 *                                                                      *
 ************************************************************************/

/************************************************************************
 *                                                                      *
 * Setting file: a setting file must be provided to configurate the     *
 *               simulated annealing.                                   *
 *                                                                      *
 ************************************************************************/

/************************************************************************
 *                                                                      *
 * Executing: simaneal <random seed> <input file> [<output directory>]  *
 *                     <output file>                                    *
 *                                                                      *
 ************************************************************************/

/************************************************************************
 *                                                                      *
 * Output: several output files are generated by SATFO.                 *
 *         1) <output file> contains information about the progress     *
 *                          of the run. Extended information about      *
 *                          function and parameter values can be        *
 *                          written to this file by setting INTERMEDIATE*
 *                          in the setting file.                        *
 *         2) Accept.out: derived acceptance ratio of single Markov     *
 *                        chain.                                        *
 *         3) Temp.out: value of control parameter of single Markov     *
 *                        chain.                                        *
 *         4) Func.out: cost function values best ever. A new value     *
 *                      is written to this file only if there is an     *
 *                      improvement.                                    *
 *         5) Mean.out: average function value in each Markov chain     *
 *         6) Heat.out: Specific heat of each markov chain              *
 *         7) Heat.out: Number of function evaluations for each chain   *
 *         8) Work.out:                                                 *
 *                                                                      *
 * If the outfile consists of two columns then the second column        *
 * represents the standard deviation                                    *
 ************************************************************************/

/************************************************************************
 *                                                                      *
 * Testing: 1) All procedures have been tested except the decrement     *
 *             schedule: DynamicGeometric.                              *
 *                                                                      *
 ************************************************************************/

/************************************************************************
 *                                                                      *
 * Length_kSkip                                                         *
 * Length_lSkip                                                         *
 * Length_jSkip                                                         *
 *                                                                      *  
 * These variables are set through the setting file. See user           *
 * documentation for additional information.                            *
 ************************************************************************/

#include "SATFOv40.h"

/**************************************************************************/
/* GLOBAL VARIABLES                                                       */
/**************************************************************************/
 int    NumberOfParameters,        /* number of parameters to optimize      */
        Accepted,                  /* Accepted transitions                  */
        Rejected,                  /* Rejected transitions                  */
        MarkovChainStep,           /* counter which gives the trial         */
                                   /* in markov chain with number:          */
        MarkovChainNumber,         /* MarkovChainNumber (counter wich       */
                                   /* gives the number of chaines           */
                                   /* generated so far                      */
        MinimumNumberTransitions,  /* Minimum number of accepted transitons */
                                   /* before markov chain is terminated     */ 
        MaximumMarkovChain,        /* Maximum number of steps in a markov   */
                                   /* chain. if exceded the chain is        */
                                   /* terminated.                           */
        MaximumRejections,         /* maximum number of rejections that     */
                                   /* are allowed to occur before markov    */
                                   /* chain is terminated.                  */
        ProposedTransitions,       /* used in heating geometric. The number */
                                   /* of transitions from which the         */
                                   /* acceptance ratio will be calculated   */ 
        MaximumNumberConstantFunctionValues, /* the number of iterations    */
                                   /* in which the function value does not  */
                                   /* change.                               */
        FunctionNumber,            /* Used to save last function values     */
        Intermediate,              /* if true prints additional output about*/
                                   /* function and parameter values to      */
                                   /* output file.                          */
        ModeHeating,               /* mechanism to heat the system          */
        ModeAccept,                /* select acceptance function            */
        ModeDecrementC,            /* select mechanism to decrease c        */
        GenerateMode,              /* select generation function            */
        ParameterOutput,           /* if true, shows results on screen      */
        COutput,                   /* if true, generates c output           */
        AROutput,                  /* if true, generates AcceptanceRatio out*/
        CFVOutput,                 /* if true, generates function values out*/
        MeanOutput,                /* if true, generates mean values out    */
        HeatOutput,                /* if true, generates heat values out    */
        EvalOutput,                /* if true, generates number evaluations */
        WorkOutput,                /* if true, generates work values        */
        NumberOfFunctionEvaluations,
        Equilibrium,               /* length markov chain to reach equilib. */
        Length_kSkip,              /* calculate equilibrium state           */
        Length_lSkip,              /* use function value for future calc.   */
        Length_jSkip;              /* calculate mean, std, spec. heat       */

 FILE   *fp;                       /* file pointer to output file           */

 char   ModeExitSA[MAXIMUM_MODES], /* mode(s) to leave SA                   */
        ModeExitMarkovChain[MAXIMUM_MODES]; /* modes to exit markov chain   */
                   
 /* dynamically allocated arrays with dimension NumberOfParameters. Except  */
 /* the array LastFunctionValues. This one has a dimension                  */
 /* MaximumNumberConstantFunctionValues                                     */
 double *Parameters,               /* Current set of optimized parameters   */
        *ParametersNew,            /* New generated parameter set           */
        *UpperBounds,              /* Upper bound for each parameter        */
        *LowerBounds,              /* Lower bound for each parameter        */
        *LastFunctionValues,       /* Last generated function values        */
        *ParametersBestEver,       /* best generated parameters ever        */
        *SpecificHeat_array;       /* specific heats during one markov chain*/

 double RandomFraction,
        c,                         /* Control parameter (temperature        */
        f_old,                     /* function value                        */
        f_new,                     /* function value                        */
        InitialAcceptanceRatio,    /* heat up until initialacceptanceratio  */
        HeatingFactorGeometric,    /* heating factor                        */
        CoolingFactorGeometric,    /* cooling factor                        */
        MinimumControlParameter,   /* minimum value of c before exit SA     */
        InitialControlParameter,   /* initial c when heating                */
        DecrementStep,             /* cooling factor                        */ 
        CoolingFactorExponential,  /* cooling factor                        */
        MinimumAcceptanceRatio,    /* minimum AR before exiting SA          */
        DistanceParameter,         /* cooling factor                        */
        StandardDeviationFunctionValue, /* stddev generated function values */
        MeanFunctionValue,         /* mean function value in markov chain   */
        MeanSquareFunctionValue,   /* <f*f>                                 */
        FunctionBestEver,          /* best generated function value ever    */
        StartingC,                 /* initial c when no heating             */
        MinimumValueCostFunction,  /* minimum value CF before exit SA       */
        FactorGeneralizedSA,       /* this factor is used with GSA          */
        OptimalValueErrorFunction, /* this values is used with GSA          */
        StepSizeParameters,        /* step size used in generative function */
        SpecificHeat;              /* Specific heat                         */
        
 char   *Directory;                /* contains filename and directory       */


/**************************************************************************/


/**************************************************************************/
/* Procedure: SA()                                                        */
/**************************************************************************/
void SA (int argc,    
	 char *argv[])
/* 
 * Author: A.H.C. van Kampen
 *         Main()
 */
{
 int    i, kSkip,lSkip,jSkip;
 int    Equilibrium, CountSpecificHeat;
 int    MeasurementCount, Global_MeasurementCount;
 int    WorkEquilCount,WorkNonEquilCount;
 double Sumx, Sumxx, Global_Sumx, Global_Sumxx, HeatDeviation;
 double WorkEquil, WorkNonEquil;
 FILE   *fp_temp, *fp_accept, *fp_func, *fp_mean, *fp_heat;
 FILE   *fp_work, *fp_eval, *fp_par;


 if(argc!=5 && argc!=4)  {
  printf("\n\nError: wrong number of arguments!\n\n");
  printf("format: simanneal <random seed> <filename input> <output directory> <filename output>\n\n");
  printf("\n         OR\n\n");
  printf("format: simanneal <random seed> <filename input> <filename output>\n\n");
  exit(1);
 }

 /*************************************************************************/
 /* Initializing simulated annealing parameters                           */
 /* These parameters are read from a setting file.                        */
 /*************************************************************************/

 Initialize(argv,
	    argc,
	    &Directory,
	    &ParametersBestEver,
	    &Parameters,
	    &ParametersNew,
	    &UpperBounds,
	    &LowerBounds,
	    &LastFunctionValues,
	    &SpecificHeat_array,
	    &NumberOfParameters,
	    &c,
	    &ModeHeating,
	    &InitialControlParameter,
	    &InitialAcceptanceRatio,
	    &HeatingFactorGeometric,
	    &ProposedTransitions,
	    &ModeDecrementC,
	    &DecrementStep,
	    &CoolingFactorExponential,
	    &CoolingFactorGeometric,
	    &DistanceParameter,
	    ModeExitSA,
	    &MinimumAcceptanceRatio,
	    &MaximumNumberConstantFunctionValues,
	    &MinimumControlParameter,
	    &MinimumValueCostFunction,
	    ModeExitMarkovChain,
	    &MinimumNumberTransitions,
	    &MaximumMarkovChain,
	    &MaximumRejections,
	    &ModeAccept,
	    &GenerateMode,
	    &Intermediate,
	    &COutput,
	    &AROutput,
	    &CFVOutput,
	    &ParameterOutput,
	    &MeanOutput,
	    &HeatOutput,
	    &EvalOutput,
            &WorkOutput,
	    &FactorGeneralizedSA,
	    &OptimalValueErrorFunction,
	    &Equilibrium,
	    &Length_kSkip,
	    &Length_lSkip,
	    &Length_jSkip,
	    &StepSizeParameters);


 /***************************************************************************/
 /* The next files are used to store the evolution of the temperature       */
 /* and acceptance ratio and the cost function values and the standard      */
 /* deviation and the mean function value for earch markov chain and the    */
 /* specific heat and Work.                                                 */
 /***************************************************************************/


 if (argc==4) strcpy(Directory,"");
 else strcpy(Directory,(char *)argv[3]);
 fp_temp=OpenOutputDevice((char *)strcat(Directory,"Temp.out"),"w",COutput);

 if (argc==4) strcpy(Directory,"");
 else strcpy(Directory,(char *)argv[3]);
 fp_accept=OpenOutputDevice((char *)strcat(Directory,"Accept.out"),"w",AROutput);

 if (argc==4) strcpy(Directory,"");
 else strcpy(Directory,(char *)argv[3]);
 fp_func=OpenOutputDevice((char *)strcat(Directory,"Func.out"),"w",CFVOutput);

 if (argc==4) strcpy(Directory,"");
 else strcpy(Directory,(char *)argv[3]);
 fp_mean=OpenOutputDevice((char *)strcat(Directory,"Mean.out"),"w",MeanOutput);

 if (argc==4) strcpy(Directory,"");
 else strcpy(Directory,(char *)argv[3]);
 fp_heat=OpenOutputDevice((char *)strcat(Directory,"Heat.out"),"w",HeatOutput);

 if (argc==4) strcpy(Directory,"");
 else strcpy(Directory,(char *)argv[3]);
 fp_eval=OpenOutputDevice((char *)strcat(Directory,"Eval.out"),"w",HeatOutput);

 if (argc==4) strcpy(Directory,"");
 else strcpy(Directory,(char *)argv[3]);
 fp_work=OpenOutputDevice((char *)strcat(Directory,"Work.out"),"w",WorkOutput);

 if (argc==4) strcpy(Directory,"");
 else strcpy(Directory,(char *)argv[3]);
 fp_par=OpenOutputDevice((char *)strcat(Directory,"Par.out"),"w",WorkOutput);


 InitializeParameters(Parameters, 
		      ParametersNew,
		      ParametersBestEver,
		      UpperBounds,
		      LowerBounds,
		      NumberOfParameters);

 PrintSAParameters(Parameters,
		   ParametersNew,
  		   NumberOfParameters,
		   UpperBounds,
		   LowerBounds,
		   c,
		   ModeHeating,
		   InitialControlParameter,
		   InitialAcceptanceRatio,
		   HeatingFactorGeometric,
		   ProposedTransitions,
		   ModeDecrementC,
		   DecrementStep,
		   CoolingFactorExponential,
		   CoolingFactorGeometric,
		   DistanceParameter,
		   ModeExitSA,
		   MinimumAcceptanceRatio,
		   MaximumNumberConstantFunctionValues,
		   MinimumControlParameter,
		   MinimumValueCostFunction,
		   ModeExitMarkovChain,
		   MinimumNumberTransitions,
		   MaximumMarkovChain,
		   MaximumRejections,
		   ModeAccept,
		   GenerateMode,
		   Intermediate,
		   COutput,
		   AROutput,
		   CFVOutput,
		   MeanOutput,
		   HeatOutput,
		   EvalOutput,
		   WorkOutput,
		   ParameterOutput,
		   FactorGeneralizedSA,
		   OptimalValueErrorFunction,
		   Equilibrium,
		   Length_kSkip,
		   Length_lSkip,
		   Length_jSkip,
		   StepSizeParameters);

 /*************************************************************************/
 /* Heating the system according to some heating schedule                 */
 /*************************************************************************/

 fprintf(fp,"\n\n************************************************\n");
 fprintf(fp,"* Heating the system.                          *\n");
 fprintf(fp,"************************************************\n");

 switch(ModeHeating) {
   case 1:
     c=InitialControlParameter;
     fprintf(fp," Heating schedule: geometrical c=c*a\n");
     fprintf(fp," Initial Acceptance ratio: %9.6lf\n",InitialAcceptanceRatio);
     fprintf(fp," Initial c: %9.6lf\n",InitialControlParameter);
     break;
   case 2:
     fprintf(fp," Heating schedule: constant c  \n");
     fprintf(fp," Final c: %9.6lf\n", c);
     break;  
 }
 Heating(ModeHeating);
 StartingC=c; /* save initial temperature */

 /*************************************************************************/
 /* Start Annealing                                                       */
 /*************************************************************************/

 fprintf(fp,"\n\n************************************************\n");
 fprintf(fp,"* Start function optimization                  *\n");
 fprintf(fp,"************************************************\n");
 
 NumberOfFunctionEvaluations=1;
 MarkovChainNumber=0;
 f_old=EvaluateCostFunction(NumberOfParameters,Parameters);

 /* Save best function value until now */
 FunctionBestEver=f_old;
 CopyVector(ParametersBestEver,Parameters,NumberOfParameters);

 lSkip=0;
 jSkip=0;

 do {  
    kSkip=0; 
    WorkEquil=0.0;
    WorkNonEquil=0.0;
    WorkEquilCount=0;
    WorkNonEquilCount=0;

    /*************************************************************************/
    /* Generate Markov chain to get the system in equilibrium                */
    /*************************************************************************/
    do { 
      kSkip++;

      SetNewParameters(GenerateMode,
		       c,
		       ParametersNew,
		       Parameters,
		       UpperBounds,
		       LowerBounds,
		       NumberOfParameters);

      f_new=EvaluateCostFunction(NumberOfParameters,ParametersNew);

      /* Save best function value and parameter set until now */
      if(f_new<FunctionBestEver) {
        FunctionBestEver=f_new;
        CopyVector(ParametersBestEver,ParametersNew,NumberOfParameters);
      }

      /***********************************************************************/
      /* New parameter set is worse than old one. Apply acceptance criterium */
      /* to see whether or not to accept this new set.                       */
      /***********************************************************************/

      if (f_new>f_old) {
        if(Accept(ModeAccept,
		  c,
		  f_new, 
		  f_old, 
		  FactorGeneralizedSA,
		  OptimalValueErrorFunction)) {
          CopyVector(Parameters, ParametersNew,NumberOfParameters);
          f_old=f_new;
          if((kSkip % Length_kSkip) == 0) {
	    WorkNonEquil += f_new;
            WorkNonEquilCount++;
          }
        }
      }
      /***********************************************************************/
      /* New parameter set is better than old one. Accept new set            */
      /***********************************************************************/
      else {
        CopyVector(Parameters,ParametersNew,NumberOfParameters);
        f_old=f_new;
	if((kSkip % Length_kSkip) == 0) {
	  WorkNonEquil -= f_new;
	  WorkNonEquilCount++;
        }
      }

    } while(kSkip<Equilibrium);
   
    /*************************************************************************/
    /* Start 'measurement' process                                           */
    /*************************************************************************/
    Rejected=0;
    Accepted=0;
    MarkovChainStep=0;
    Sumx=0.0;
    Sumxx=0.0;
    Global_Sumx=0.0;
    Global_Sumxx=0.0;
    FunctionNumber=0;
    CountSpecificHeat=0;
    MeasurementCount=0;
    Global_MeasurementCount=0;

    ++MarkovChainNumber;

    /*************************************************************************/
    /* Generate Markov chain                                                 */
    /*************************************************************************/

    do { 
      lSkip++;
      jSkip++;

      ++MarkovChainStep;

      SetNewParameters(GenerateMode,
		       c,
		       ParametersNew,
		       Parameters,
		       UpperBounds,
		       LowerBounds,
		       NumberOfParameters);

      f_new=EvaluateCostFunction(NumberOfParameters,ParametersNew);
      ++NumberOfFunctionEvaluations;

      /* Save best function value and parameter set until now */
      if(f_new<FunctionBestEver) {
        FunctionBestEver=f_new;
        CopyVector(ParametersBestEver,ParametersNew,NumberOfParameters);
      }

      /* Save last function value */
      /* only a number of previous function values are saved  */
      /* if we have saved this number of function values      */
      /* the array is overwritten with succeeding values      */
      /* starting with the first one in the array until       */
      /* we have again filled the array.                      */
      LastFunctionValues[FunctionNumber]=f_old;    
      if(++FunctionNumber==MaximumNumberConstantFunctionValues)
         FunctionNumber=0;

      /***********************************************************************/
      /* New parameter set is worse than old one. Apply acceptance criterium */
      /* to see whether or not to accept this new set.                       */
      /***********************************************************************/

      if (f_new>f_old) {
        if(Accept(ModeAccept,
		  c,
		  f_new, 
		  f_old, 
		  FactorGeneralizedSA,
		  OptimalValueErrorFunction)) {
          CopyVector(Parameters, ParametersNew,NumberOfParameters);
          f_old=f_new;
          if((lSkip % Length_lSkip)==0) ++Accepted;
          if((kSkip % Length_kSkip) == 0) {
	    WorkEquil += f_new;
	    WorkEquilCount++;
          }
        }
	else {
	  if((lSkip % Length_lSkip)==0) ++Rejected;
        }
      }
      /***********************************************************************/
      /* New parameter set is better than old one. Accept new set            */
      /***********************************************************************/
      else {
        if((lSkip % Length_lSkip)==0) ++Accepted;
        CopyVector(Parameters,ParametersNew,NumberOfParameters);
        f_old=f_new;
        if((kSkip % Length_kSkip) == 0) {
	  WorkEquil -= f_new;
	  WorkEquilCount++;
        }
      }
 
     if ((lSkip % Length_lSkip)==0) {
	Sumx += f_old;
	Sumxx += f_old*f_old;
	Global_Sumx += f_old;
	Global_Sumxx += f_old*f_old;
        MeasurementCount++;
        Global_MeasurementCount++;
      }

      if((jSkip % Length_jSkip)==0) {
	/*********************************************************************/
	/* Calculate mean and standard deviation of generated function values*/
	/*********************************************************************/

        MeanSquareFunctionValue = Sumxx/MeasurementCount;
        MeanFunctionValue       = Sumx/MeasurementCount;
        SpecificHeat_array[CountSpecificHeat]=
              (MeanSquareFunctionValue-SQR(MeanFunctionValue))/SQR(c);

        CountSpecificHeat+=1;      
        Sumx=0.0;
        Sumxx=0.0;
        MeasurementCount=0;
      }

      /***********************************************************************/
      /* Print new parameter set and function value if intermediate=TRUE     */
      /***********************************************************************/
      if(Intermediate) {
        for(i=0;i<NumberOfParameters;i++) 
           if(i==0)
	     fprintf(fp,"\n P%d: %lf         f: %lf\n",i,Parameters[i],f_old);
           else
	     fprintf(fp," P%d: %lf\n",i,Parameters[i]);
      }

      fflush(fp); /* write data to file */
   
    } while(!ExitMarkovChain(ModeExitMarkovChain));

    WorkNonEquil /= WorkNonEquilCount;
    WorkEquil /= WorkEquilCount;

    /*************************************************************************/
    /* Save results of one markov chain                                      */
    /*************************************************************************/
    fprintf(fp," \n\n=========================================================\n");
    fprintf(fp," Markov chain number: %d\n", MarkovChainNumber);
    fprintf(fp," c:%g\n",c);
    fprintf(fp," Acceptance ratio: %g\n", 
                  AcceptanceRatio(Global_MeasurementCount,Accepted));
    fprintf(fp," Length markov chain: %d  Number of transitions: %d\n",
	            MarkovChainStep,
	            Accepted);
    fprintf(fp," Number of function evaluations: %d\n",
	           NumberOfFunctionEvaluations);

    /* print cost function best ever value to file */
    fprintf(fp_func,"%lf\n",FunctionBestEver);

    /* save c and acceptance ratio in seperate files */
    fprintf(fp_temp,"%g\n",c);
    fprintf(fp_accept,"%g\n",AcceptanceRatio(Global_MeasurementCount,Accepted));
    fprintf(fp_work,"%g %g \n",WorkNonEquil, WorkEquil);    

    /*************************************************************************/
    /* Calculate mean and standard deviation of generated function values    */
    /*************************************************************************/
  

    fprintf(fp,"\n\n************************************************\n");
    fprintf(fp,"* Calculate Global mean and standard deviation *\n");
    fprintf(fp,"* of generated function values.                *\n");
    fprintf(fp,"************************************************\n");

    MeanFunctionValue = Global_Sumx/Global_MeasurementCount;
    StandardDeviationFunctionValue= 
             sqrt((Global_Sumxx-SQR(Global_Sumx)/Global_MeasurementCount) /
                  (Global_MeasurementCount-1));
        
    /*************************************************************************/
    /* Calculate mean specific heat and deviation specific heat              */
    /*************************************************************************/

    SpecificHeat=0.0;
    Sumx=0.0;
    Sumxx=0.0;
    for(i=0;i<CountSpecificHeat;i++) {
      Sumx += SpecificHeat_array[i];
      Sumxx += SQR(SpecificHeat_array[i]);
      SpecificHeat+=SpecificHeat_array[i];
    }
    SpecificHeat /= (CountSpecificHeat); /* from ZERO to CountSpecificHeat */
    HeatDeviation=sqrt((Sumxx-SQR(Sumx)/CountSpecificHeat)/(CountSpecificHeat-1));

    /* save mean, std_dev and heat in seperate files */
    fprintf(fp_mean,"%lf %lf\n", MeanFunctionValue,
                                 StandardDeviationFunctionValue);
    fprintf(fp_heat,"%lf %lf\n", SpecificHeat, HeatDeviation);
    fprintf(fp," Mean: %lf, Standard deviation %lf\n",
                 MeanFunctionValue, StandardDeviationFunctionValue);

    
    /*************************************************************************/
    /* Decrement control parameter according to some cooling schedule       */
    /*************************************************************************/
   
    fprintf(fp,"\n\n************************************************\n");
    fprintf(fp,"* Cooling the system.                          *\n");
    fprintf(fp,"************************************************\n");

    switch(ModeDecrementC) {
      case 1:
        fprintf(fp," Cooling schedule: Geometric\n");
        fprintf(fp," Cooling factor geometric: %lf\n",CoolingFactorGeometric);
        break;
      case 2:
        fprintf(fp," Cooling schedule: Dynamic geometric\n");
        fprintf(fp," Distance parameter: %lf\n",DistanceParameter);
	fprintf(fp," Standard deviation: %lf\n",StandardDeviationFunctionValue);
        break;
      case 3:
        fprintf(fp," Cooling schedule: linear\n");
        fprintf(fp," Decrement step: %lf\n",DecrementStep); 
        break;
      case 4:
        fprintf(fp," Cooling schedule: Geometric2\n");
        fprintf(fp," Cooling factor (markov chain number): %d\n",
                       MarkovChainNumber);

        break;
      case 5:
        fprintf(fp," Cooling schedule: Logarithmic\n");
        fprintf(fp," Cooling factor (markov chain number): %d\n",
                       MarkovChainNumber);
        break;
      case 6:
        fprintf(fp," Cooling schedule: Exponential\n");
        fprintf(fp," Cooling factor exponential: %lf\n",
                     CoolingFactorExponential);
        fprintf(fp," Markov chain number %d\n", MarkovChainNumber);
    }

    DecrementControlParameter(ModeDecrementC);

    if(ParameterOutput==1) {
      printf("\n\n Markov chain number: %d\n", MarkovChainNumber);
      printf(" c:%lf\n",c);
      printf(" Acceptance ratio: %lf\n", 
                    AcceptanceRatio(Global_MeasurementCount,Accepted));
      printf(" Length markov chain: %d  Number of transitions: %d\n",
                    MarkovChainStep, Accepted);
      printf(" Number of function evaluations: %d\n",
                  NumberOfFunctionEvaluations);

      fprintf(fp_eval,"%d %d %d %d\n",MarkovChainNumber,
	                              MarkovChainStep,
	                              Accepted,
                                      NumberOfFunctionEvaluations);

      printf("\n");
      printf(" Best %lf  Now %lf\n",FunctionBestEver,f_old);

      if (argc==4) strcpy(Directory,"");
      else strcpy(Directory,(char *)argv[3]);
      fp_par=OpenOutputDevice((char *)strcat(Directory,"Par.out"),"w",WorkOutput);

      fprintf(fp_par," Best %lf  Now %lf\n",FunctionBestEver,f_old);
      fprintf(fp_par," ------------------------------------\n");
      for(i=0;i<NumberOfParameters;i++) 
	fprintf(fp_par," Best%d: %10.7lf  Now%d: %10.7lf\n",
		i,ParametersBestEver[i],
		i,Parameters[i]);
    fflush(fp_par);
    fclose(fp_par);
    }

    fflush(fp_mean);
    fflush(fp_heat);
    fflush(fp_temp);
    fflush(fp_accept);
    fflush(fp_func);
    fflush(fp_eval);
    fflush(fp_work);

 } while(!ExitSimulatedAnnealing(ModeExitSA));

 fprintf(fp,"\n\n************************************************\n");
 fprintf(fp,"* Final parameters and function values         *\n");
 fprintf(fp,"************************************************\n");

 for(i=0;i<NumberOfParameters;i++) 
    if(i==0)
      fprintf(fp," P%d: %lf          f: %lf\n",i,Parameters[i],f_old);
    else
      fprintf(fp," P%d: %lf\n",i,Parameters[i]);

 fprintf(fp,"\n\n************************************************\n");
 fprintf(fp,"* Parameters best ever                         *\n");
 fprintf(fp,"************************************************\n");

 for(i=0;i<NumberOfParameters;i++) 
    if(i==0)
      fprintf(fp," P%d: %lf          f: %lf\n",i,ParametersBestEver[i],
	                                       FunctionBestEver);
    else
      fprintf(fp," P%d: %lf\n",i,ParametersBestEver[i]);


 FREE1D(ParametersBestEver); 
 FREE1D(Parameters);
 FREE1D(ParametersNew);
 FREE1D(LastFunctionValues);
 FREE1D(UpperBounds);
 FREE1D(LowerBounds);
 FREE1D(SpecificHeat_array);

 fclose(fp);
 fclose(fp_temp);
 fclose(fp_accept);
 fclose(fp_func);
 fclose(fp_mean);
 fclose(fp_heat);
 fclose(fp_eval);
 fclose(fp_par);
 fclose(fp_work);
}


/**************************************************************************/
/**                                                                      **/
/**                                                                      **/
/**                           INITIALIZATION                             **/
/**                                                                      **/
/**                                                                      **/
/**************************************************************************/

/**************************************************************************/
/* Procedure: Initialize                                                  */
/**************************************************************************/
void Initialize(char   *argv[],
		int    argc,
		char   **Directory,
		double **ParametersBestEver,
		double **Parameters,
		double **ParametersNew,
		double **UpperBounds,
		double **LowerBounds,
		double **LastFunctionValues,
                double **SpecificHeat_array,
                int    *NumberOfParameters,
                double *c,
		int    *ModeHeating,
		double *InitialControlParameter,
		double *InitialAcceptanceRatio,
		double *HeatingFactorGeometric,
		int    *ProposedTransitions,
		int    *ModeDecrementC,
		double *DecrementStep,
		double *CoolingFactorExponential,
		double *CoolingFactorGeometric,
		double *DistanceParameter,
		char   *ModeExitSA,
		double *MinimumAcceptanceRatio,
		int    *MaximumNumberConstantFunctionValues,
		double *MinimumControlParameter,
		double *MinimumValueCostFunction,
		char   *ModeExitMarkovChain,
		int    *MinimumNumberTransitions,
		int    *MaximumMarkovChain,
		int    *MaximumRejections,
		int    *ModeAccept,
		int    *GenerateMode,
		int    *Intermediate,
		int    *COutput,
		int    *AROutput,
		int    *CFVOutput,
		int    *ParameterOutput,
		int    *MeanOutput,
		int    *HeatOutput,
		int    *EvalOutput,
		int    *WorkOutput,
		double *FactorGeneralizedSA,
		double *OptimalValueErrorFunction,
		int    *Equilibrium,
		int    *Length_kSkip,
		int    *Length_lSkip,
		int    *Length_jSkip,
		double *StepSizeParameters)
/*
 * Author: A.H.C. van Kampen, 16-7-93
 * Purpose: initialization of variables
 *          The file .set is read in this procedure
 */

{
 long int RandomSeed;
 int    i, IntegerValue, MainOutput;
 double RealValue;
 char   StringValue[MAXIMUM_MODES];
 FILE   *fp2;


 /* initialize random generator */ 
 sscanf(argv[1],"%lu",&RandomSeed); 
 SA_RMARIN(RandomSeed);

 /* input filename: setting file */
 fp2=OpenFile(argv[2],"r");

 FScanColon(fp2);
 fscanf(fp2,"%d",&IntegerValue);
 *NumberOfParameters=IntegerValue;

 CALLOC1D(*ParametersBestEver,*NumberOfParameters);
 CALLOC1D(*Parameters,*NumberOfParameters);
 CALLOC1D(*ParametersNew,*NumberOfParameters);
 CALLOC1D(*UpperBounds,*NumberOfParameters);
 CALLOC1D(*LowerBounds,*NumberOfParameters);

 FScanColon(fp2);
 for(i=0;i<*NumberOfParameters;i++) {
    fscanf(fp2,"%lf",&RealValue);
    (*LowerBounds)[i]=RealValue;
 }

 FScanColon(fp2);
 for(i=0;i<*NumberOfParameters;i++) {
    fscanf(fp2,"%lf",&RealValue);
    (*UpperBounds)[i]=RealValue;
 }

 FScanColon(fp2);
 fscanf(fp2,"%d",&IntegerValue);
 *GenerateMode=IntegerValue;

 FScanColon(fp2);
 fscanf(fp2,"%lf",&RealValue);
 *StepSizeParameters=RealValue;

 FScanColon(fp2); 
 fscanf(fp2,"%d",&IntegerValue);
 *ModeHeating=IntegerValue;

 FScanColon(fp2);
 fscanf(fp2,"%lf",&RealValue);
 *InitialControlParameter=RealValue;

 FScanColon(fp2);
 fscanf(fp2,"%lf",&RealValue);
 *HeatingFactorGeometric=RealValue;

 FScanColon(fp2);
 fscanf(fp2,"%lf",&RealValue);
 *InitialAcceptanceRatio=RealValue;

 FScanColon(fp2);
 fscanf(fp2,"%d",&IntegerValue);
 *ProposedTransitions=IntegerValue;

 FScanColon(fp2);
 fscanf(fp2,"%lf",&RealValue);
 *c=RealValue;

 FScanColon(fp2); 
 fscanf(fp2,"%d",&IntegerValue);
 *ModeDecrementC=IntegerValue;

 FScanColon(fp2);
 fscanf(fp2,"%lf",&RealValue);
 *CoolingFactorGeometric=RealValue;

 FScanColon(fp2);
 fscanf(fp2,"%lf",&RealValue);
 *CoolingFactorExponential=RealValue;

 FScanColon(fp2);
 fscanf(fp2,"%lf",&RealValue);
 *DecrementStep=RealValue;

 FScanColon(fp2);
 fscanf(fp2,"%lf",&RealValue);
 *DistanceParameter=RealValue;

 FScanColon(fp2); 
 fscanf(fp2,"%d",&IntegerValue);
 *Equilibrium=IntegerValue;

 FScanColon(fp2); 
 fscanf(fp2,"%s",&StringValue);
 strcpy(ModeExitMarkovChain,StringValue);

 FScanColon(fp2); 
 fscanf(fp2,"%d",&IntegerValue);
 *MinimumNumberTransitions=IntegerValue;

 FScanColon(fp2); 
 fscanf(fp2,"%d",&IntegerValue);
 *MaximumMarkovChain=IntegerValue;

 FScanColon(fp2); 
 fscanf(fp2,"%d",&IntegerValue);
 *MaximumNumberConstantFunctionValues=IntegerValue;

 FScanColon(fp2); 
 fscanf(fp2,"%d",&IntegerValue);
 *MaximumRejections=IntegerValue;

 CALLOC1D(*LastFunctionValues,*MaximumNumberConstantFunctionValues);

 /* initialize LastFunctionValues */
 for(i=0;i<*MaximumNumberConstantFunctionValues;i++)
    (*LastFunctionValues)[i]=(double)i;

 FScanColon(fp2); 
 fscanf(fp2,"%s",&StringValue);
 strcpy(ModeExitSA,StringValue);

 FScanColon(fp2);
 fscanf(fp2,"%lf",&RealValue);
 *MinimumControlParameter=RealValue;

 FScanColon(fp2);
 fscanf(fp2,"%lf",&RealValue);
 *MinimumAcceptanceRatio=RealValue;

 FScanColon(fp2);
 fscanf(fp2,"%lf",&RealValue);
 *MinimumValueCostFunction=RealValue;

 FScanColon(fp2); 
 fscanf(fp2,"%d",&IntegerValue);
 *ModeAccept=IntegerValue;

 FScanColon(fp2); 
 fscanf(fp2,"%lf",&RealValue);
 *OptimalValueErrorFunction=RealValue;

 FScanColon(fp2); 
 fscanf(fp2,"%lf",&RealValue);
 *FactorGeneralizedSA=RealValue;

 FScanColon(fp2); 
 fscanf(fp2,"%d",&IntegerValue);
 *Length_kSkip=IntegerValue;

 FScanColon(fp2); 
 fscanf(fp2,"%d",&IntegerValue);
 *Length_lSkip=IntegerValue;

 FScanColon(fp2); 
 fscanf(fp2,"%d",&IntegerValue);
 *Length_jSkip=IntegerValue;

 FScanColon(fp2); 
 fscanf(fp2,"%d",&IntegerValue);
 MainOutput=IntegerValue;


 CALLOC1D(*Directory, MAXIMUM_LENGTH_FILENAME);
 if (argc==4) 
   strcpy(*Directory,(char *)argv[3]); 
 else {
   strcpy(*Directory,(char *)argv[3]);
   strcat(*Directory,(char *)argv[4]);
 }
 fp=OpenOutputDevice(*Directory,"w",MainOutput);

 fprintf(fp,"\n\n************************************************\n");
 fprintf(fp,"* Initializing simulated annealing parameters  *\n");
 fprintf(fp,"************************************************\n");


 FScanColon(fp2); 
 fscanf(fp2,"%d",&IntegerValue);
 *Intermediate=IntegerValue;

 FScanColon(fp2); 
 fscanf(fp2,"%d",&IntegerValue);
 *COutput=IntegerValue;

 FScanColon(fp2); 
 fscanf(fp2,"%d",&IntegerValue);
 *AROutput=IntegerValue;

 FScanColon(fp2); 
 fscanf(fp2,"%d",&IntegerValue);
 *CFVOutput=IntegerValue;

 FScanColon(fp2); 
 fscanf(fp2,"%d",&IntegerValue);
 *ParameterOutput=IntegerValue;

 FScanColon(fp2); 
 fscanf(fp2,"%d",&IntegerValue);
 *MeanOutput=IntegerValue;

 FScanColon(fp2); 
 fscanf(fp2,"%d",&IntegerValue);
 *HeatOutput=IntegerValue;

 FScanColon(fp2); 
 fscanf(fp2,"%d",&IntegerValue);
 *EvalOutput=IntegerValue;

 FScanColon(fp2); 
 fscanf(fp2,"%d",&IntegerValue);
 *WorkOutput=IntegerValue;

 CALLOC1D(*SpecificHeat_array,*MaximumMarkovChain);

 close(fp2);
}

/**************************************************************************/
/* Procedure: InitializeParameters()                                      */
/**************************************************************************/
void InitializeParameters(double *Parameters,          
			  double *ParametersNew,       
			  double *ParametersBestEver,  
			  double *UpperBounds,         
			  double *LowerBounds,         
			  int     NumberOfParameters)  
/*
 * Author: A.H.C. van Kampen, 26-7-93
 * Purpose: initializes the arrays *Parameters(New)(BestEver) with
 *          random values between the constraints for each parameter, i.e.
 *            *UpperBounds, *LowerBounds
 */
{
 int i;

 for(i=0;i<NumberOfParameters;i++) {
    Parameters[i]=
    ParametersNew[i]=
    ParametersBestEver[i]=
      (UpperBounds[i]-LowerBounds[i])*RandomSA(&RandomFraction)+LowerBounds[i];
 }
}


/**************************************************************************/
/**                                                                      **/
/**                                                                      **/
/**                           GENERAL FUNCTIONS                          **/
/**                                                                      **/
/**                                                                      **/
/**************************************************************************/


/**************************************************************************/
/* Procedure: OpenFile                                                    */
/**************************************************************************/
FILE *OpenFile (char *FileName, /* R-/1C */
		char *Mode)     /* R-/0C */
/* 
 * Author: A.H.C. van Kampen, 12-5-93
 * Purpose: Opens the file *Filename in mode *Mode 
 */
{
 FILE *fp;

 if (!(fp=fopen(FileName, Mode))) {
   printf("Error in OpenFile().\n");
   printf("Not able to open file %s, mode %s.\n",FileName,Mode);
   printf("Exiting.... \n\n");
   exit(1);
 }
 
 return fp;
}

/**************************************************************************/
/* Procedure: OpenOutputDevice()                                          */
/**************************************************************************/
FILE *OpenOutputDevice(char *FileName,
		       char *mode,     
		       int   Device)   
/*
 * Author: A.H.C. van Kampen, 28-7-93
 * Purpose: redirects output to Device (File,Screen or no output)
 *          if output to file, then open file *Filename in mode *mode
 */
{
 FILE *Out;

 switch(Device) {
   case TO_FILE:     Out=OpenFile(FileName,mode);
     break;
   case TO_SCREEN:
     Out=stderr;
     break;
   case NO_OUTPUT:
     Out=OpenFile(DEV_NULL,mode);
     break;
 }
 return(Out);
}

/**************************************************************************/
/* Procedure: FScanColon()                                                */
/**************************************************************************/
void FScanColon(FILE *fp) 
/*
 * Author: C.B. Lucasius
 * Purpose: scans file pointer by *fp until colon is encounterd
 */
{
 fscanf(fp,"%*[^:]");
 fscanf(fp,": ");
}

/**************************************************************************/
/* Procedure: RandomSA()                                                  */
/**************************************************************************/
double RandomSA(double *RandomFraction) 
/*
 * Author: A.H.C. van Kampen, 22-7-93
 * Purpose: returns random value between [0,1)
 */
{
 return(fabs(RANDOM_BOUND_DOUBLE(1,RandomFraction)));
}

/**************************************************************************/
/* Procedure: CopyVector()                                                */
/**************************************************************************/
void CopyVector(double *Vector1,   
		double *Vector2,    
		int     Dimension)  
/*
 * Author: A.H.C. van Kampen, 14-7-93
 * Purpose: copies one double array into a second array
 */
{
 int i;

 for(i=0;i<Dimension;i++) 	
    Vector1[i]=Vector2[i];
}

/**************************************************************************/
/* Procedure: CheckMode()                                                 */
/**************************************************************************/
int CheckMode(int Mode,           
	      char *ModeString)   
/*
 * Author: A.H.C. van Kampen, 15-7-93
 * Purpose: Check if Mode occurs in *ModeString
 */
{
 int  i;
 char p[2];

 sprintf(p,"%d",Mode);
 for(i=0;i<strlen(ModeString);i++) 
   if(ModeString[i]==p[0]) return(TRUE);
 
 return(FALSE);

}

/**************************************************************************/
/* Procedure: PrintSAParameters()                                         */
/**************************************************************************/
void PrintSAParameters(double *Parameters,
		       double *ParametersNew,
		       int    NumberOfParameters,
		       double *UpperBounds,
		       double *LowerBounds,
		       double c,
		       int    ModeHeating,
		       double InitialControlParameter,
		       double InitialAcceptanceRatio,
		       double HeatingFactorGeometric,
		       int    ProposedTransitions,
		       int    ModeDecrementC,
		       double DecrementStep,
		       double CoolingFactorExponential,
		       double CoolingFactorGeometric,
		       double DistanceParameter,
		       char   *ModeExitSA,
		       double MinimumAcceptanceRatio,
		       int    MaximumNumberConstantFunctionValues,
		       double MinimumControlParameter,
		       double MinimumValueCostFunction,
		       char   *ModeExitMarkovChain,
		       int    MinimumNumberTransitions,
		       int    MaximumMarkovChain,
		       int    MaximumRejections,
		       int    ModeAccept,
		       int    GenerateMode,
		       int    Intermediate,
		       int    COutput,
		       int    AROutput,
		       int    CFVOutput,
		       int    ParameterOutput,
		       int    MeanOutput,
		       int    HeatOutput,
		       int    EvalOutput,
		       int    WorkOutput,
		       double FactorGeneralizedSA,
		       double OptimalValueErrorFunction,
		       int    Equilibrium,
		       int    Length_kSkip,
		       int    Length_lSkip,
		       int    Length_jSkip,
		       double StepSizeParameters)
/*
 * Author: A.H.C. van Kampen, 15-7-93
 * Purpose: Print global variables
 */
{
 int i;

 fprintf(fp,"\n\n************************************************\n");
 fprintf(fp,"* Upper/Lower bounds                           *\n");
 fprintf(fp,"************************************************\n");

 fprintf(fp,"NumberOfParameters %d\n",NumberOfParameters);
 for(i=0;i<NumberOfParameters;i++)
  fprintf(fp,"Upper: %lf Lower: %lf \n",
	  UpperBounds[i], LowerBounds[i]);

 fprintf(fp,"\n\n************************************************\n");
 fprintf(fp,"* Initial parameters                           *\n");
 fprintf(fp,"************************************************\n");

 for(i=0;i<NumberOfParameters;i++)
  fprintf(fp,"P: %lf\n",Parameters[i]);

 fprintf(fp,"\nc: %lf\n",c);
 fprintf(fp,"Mode heating: %d\n",ModeHeating);
 fprintf(fp,"InitialControlParameter %lf\n",InitialControlParameter);
 fprintf(fp,"InitialAcceptanceRatio %lf\n",InitialAcceptanceRatio);
 fprintf(fp,"HeatingFactorGeometric %lf\n",HeatingFactorGeometric);
 fprintf(fp,"ProposedTransitions %d\n",ProposedTransitions);
 fprintf(fp,"ModeDecrementC %d\n",ModeDecrementC);
 fprintf(fp,"DecrementStep %lf\n",DecrementStep);
 fprintf(fp,"CoolingFactorExponential %lf\n",CoolingFactorExponential);
 fprintf(fp,"CoolingFactorGeometric %lf\n",CoolingFactorGeometric);
 fprintf(fp,"DistanceParameter %lf\n",DistanceParameter);
 fprintf(fp,"ModeExitSA %s\n", ModeExitSA);
 fprintf(fp,"MinimumAcceptanceRatio %lf\n",MinimumAcceptanceRatio);
 fprintf(fp,"MaximumNumberCFV: %d\n",MaximumNumberConstantFunctionValues);
 fprintf(fp,"MinimumControlParameter %lf\n",MinimumControlParameter);
 fprintf(fp,"MinimumValueCostFunction %lf\n",MinimumValueCostFunction);
 fprintf(fp,"ModeExitMarkovChain %s\n",ModeExitMarkovChain);
 fprintf(fp,"MinimumNumberTransitions %d\n",MinimumNumberTransitions);
 fprintf(fp,"MaximumMarkovChain %d\n",MaximumMarkovChain);
 fprintf(fp,"MaximumRejections %d\n",MaximumRejections);
 fprintf(fp,"ModeAccept %d\n",ModeAccept);
 fprintf(fp,"GenerateMode %d\n",GenerateMode);
 fprintf(fp,"Intermediate %d\n",Intermediate);
 fprintf(fp,"Control parameter Output %d\n",COutput);
 fprintf(fp,"Acceptance ratio output %d\n",AROutput);
 fprintf(fp,"CostFunctionValues output %d\n",CFVOutput);
 fprintf(fp,"ParameterOutput %d\n",ParameterOutput);
 fprintf(fp,"MeanOutput %d\n",MeanOutput);
 fprintf(fp,"HeatOutput %d\n",HeatOutput);
 fprintf(fp,"HeatOutput %d\n",WorkOutput);
 fprintf(fp,"EvaluationOutput %d\n",EvalOutput);
 fprintf(fp,"FactorGeneralizedSA %lf\n",FactorGeneralizedSA);
 fprintf(fp,"OptimalValueErrorFunction %lf \n",OptimalValueErrorFunction);
 fprintf(fp,"Equilibrium %d\n",Equilibrium);
 fprintf(fp,"Skip equilibrium %d\n",Length_kSkip);
 fprintf(fp,"Skip function values %d\n",Length_lSkip);
 fprintf(fp,"Skip Specific Heat %d\n",Length_jSkip);
 fprintf(fp,"Step size parameters %lf\n\n\n",StepSizeParameters);
}


/**************************************************************************/
/**                                                                      **/
/**                                                                      **/
/**                       C DECREMENT SCHEDULES                          **/
/**                                                                      **/
/**                                                                      **/
/**************************************************************************/


/**************************************************************************/
/* Procedure: DecrementControlParameter()                                 */
/**************************************************************************/
void DecrementControlParameter(int Mode)
/*
 * Author: A.H.C. van Kampen, 26-7-93
 * Purpose:  determine which schedule to use for temperature decrement
 */
{
 switch(Mode) {
 case 1:
   DecrementControlParameterGeometric(&c,
				      CoolingFactorGeometric);
   break;
 
 case 2:
   DecrementControlParameterDynamicGeometric(&c,
					     DistanceParameter,
					     StandardDeviationFunctionValue);
   break;

 case 3:
   DecrementControlParameterLinear(&c,
				   DecrementStep);
   break;

 case 4: 
   DecrementControlParameterGeometric2(&c,
				       StartingC,
				       MarkovChainNumber);
   break;
 
 case 5:
   DecrementControlParameterLogarithmic(&c,
					StartingC,
					MarkovChainNumber);
   break;
 
 case 6:
   DecrementControlParameterExponential(&c,
					StartingC,
					CoolingFactorExponential,
					MarkovChainNumber);
   break;
 }
}

/**************************************************************************/
/* Procedure: DecrementControlParameterDynamicGeometric()                 */
/**************************************************************************/
void DecrementControlParameterDynamicGeometric(double *c,
					       double DistanceParameter,
					       double StandardDeviation)
/*
 * Author: A.H.C. van Kampen, 27-7-93
 * Purpose:  DecrementControlParameterDynamicGeometric
 *           Laarhoven, P.J.M. and Aarts, E.H.L., 'simulated annealing     
 *           theory and applications', Reidel, Dordrecht, 1987 
 */
{
 *c = *c / (1.0+(*c*log(1.0+DistanceParameter)/3.0*StandardDeviation));
}

/**************************************************************************/
/* Procedure: DecrementControlParameterLinear                             */
/**************************************************************************/
void DecrementControlParameterLinear(double *c,
				     double DecrementStep)
/*
 * Author: A.H.C. van Kampen, 27-7-93
 * Purpose: DecrementControlParameterLinear
 *          Barakat, M.T., Journal Comp-aided mol design, 4, 295-316
 */
{
 *c = *c - DecrementStep;

 if(*c<=0.0) {
   fprintf(fp," Warning decrement c: temperature negative or zero!\n");
   fprintf(fp," Simulated annealing is exited. Adjust decrement steps\n");
   fprintf(fp," or mode to exit simulated annealing.\n");

   fprintf(stderr," Warning decrement c: temperature negative or zero!\n");
   fprintf(stderr," Simulated annealing is exited. Adjust decrement steps\n");
   fprintf(stderr," or mode to exit simulated annealing.\n");

   close(fp);
   exit(1);
 }
}

/**************************************************************************/
/* Procedure: DecrementControlParameterLogarithmic()                      */
/**************************************************************************/
void DecrementControlParameterLogarithmic(double *c,
					  double  c0,
					  int     MarkovChainNumber)
/*
 * Author: A.H.C. van Kampen, 27-7-93
 * Purpose: DecrementControlParameterLogarithmic
 */
{
 *c = c0 / log(1.0+(double)MarkovChainNumber);
}

/**************************************************************************/
/* Procedure: DecrementControlParameterExponentail()                      */
/**************************************************************************/
void DecrementControlParameterExponential(double *c,
					  double  c0,
					  double  CoolingFactorExponential,
					  int     MarkovChainNumber)
/*
 * Author: A.H.C. van Kampen, 27-7-93
 * Purpose: DecrementControlParameterExponential
 */
{
 *c = c0 * pow(CoolingFactorExponential, (double)MarkovChainNumber);
}

/**************************************************************************/
/* Procedure: DecrementControlParameterGeometric()                        */
/**************************************************************************/
void DecrementControlParameterGeometric(double *c,
					double  CoolingFactor)
/*
 * Author: A.H.C. van Kampen, 14-7-93
 * Purpose:  DecrementControlParameterGeometric
 *           Laarhoven, P.J.M. and Aarts, E.H.L., 'simulated annealing     
 *           theory and applications', Reidel, Dordrecht, 1987 
 */
{
 *c = CoolingFactor * *c;
}

/**************************************************************************/
/* Procedure: DecrementControlParameterGeometric2                         */
/**************************************************************************/
void DecrementControlParameterGeometric2(double *c,
					 double  c0,
					 int     MarkovChainNumber)
/*
 * Author: A.H.C. van Kampen, 27-7-93
 * Purpose:  DecrementControlParameterGeometric2
 */
{
 *c = c0 / (1.0 + (double)MarkovChainNumber);
}


/**************************************************************************/
/**                                                                      **/
/**                                                                      **/
/**                        C INCREMENT SCHEDULES                         **/
/**                                                                      **/
/**                                                                      **/
/**************************************************************************/


/**************************************************************************/
/* Procedure: Heating()                                                   */
/**************************************************************************/
void Heating(int Mode)
/*
 * Author: A.H.C. van Kampen, 1-9-93
 * Purpose:  determines which heating schedule is incooperated 
 *           ModeString: string which contains one decimal value <=9,>=0
 */
{
 switch(Mode) {
   case 1:
    HeatingGeometric(&c, 
		     Parameters,
		     NumberOfParameters,
		     InitialAcceptanceRatio,
		     HeatingFactorGeometric,
		     ProposedTransitions,
		     ModeAccept,
		     GenerateMode,
		     FactorGeneralizedSA,
		     OptimalValueErrorFunction);
    break;

   case 2:
    HeatingSetConstant(c);
    break;
  }
}

/**************************************************************************/
/* Procedure: HeatingSetConstant()                                        */
/**************************************************************************/
void HeatingSetConstant(double c)
/*
 * Author: A.H.C. van Kampen, 22-7-93
 * Purpose: sets control parameter to user defined initial value  
 */
{
 /* *c has already its correct value because it was directly read */
 /* from SA.set                                                   */
}

/**************************************************************************/
/* Procedure: HeatingGeometric                                            */
/**************************************************************************/
void HeatingGeometric(double *c,                        
		      double *Parameters,               
		      int     NumberOfParameters,       
		      double  InitialAcceptanceRatio,   
		      double  HeatingFactor,            
		      int     ProposedTransitions,      
		      int     ModeAccept,               
		      int     GenerateMode,             
		      double  FactorGeneralizedSA,
		      double  OptimalValueErrorFunction)
/*
 * Author: A.H.C. van Kampen, 1-9-93
 * Purpose:  sets initial value for control parameter 'c' according to    
 *           the simple criterium that the AcceptanceRatio (which is      
 *           defined as (accepted transitions)/(proposed transitions) in  
 *           the case that the new function value f1 is less than the     
 *           old function value f2), must be bigger  than some predefined 
 *           value (InitialAcceptanceRatio). This can be accomplished by  
 *           starting at a sufficiently low value for c (for example 1.0) 
 *           and increasing this value until the MinimumAcceptance is     
 *           reached.                                                     
 *           The increase in temperature is in this case defined by       
 *                 c = c * HeatingFractor
 *
 *          KirkPatrick, S., 'Optimization by simulated annealing, IBM    
 *          research report RC9355, 1982.
 */
{
 int    Proposed, Accepted;
 double	f1, f2, *ParametersTemp, *ParametersNew, InitialFunctionValue;

 CALLOC1D(ParametersTemp,NumberOfParameters);
 CALLOC1D(ParametersNew,NumberOfParameters);

 /* prevent that c has the illegal value 0 */
 if (*c < MACHINE_PRECISION) {
   fprintf(fp,"\n\nWarning: in trying to heat the system, the \n");
   fprintf(fp,"control parameter c has the value %lf\n",*c);
   fprintf(fp,"This is a illegal value. c is set to 1.0.\n\n");
   *c = 1.0;
 }

 /* save initial parameters */
 CopyVector(ParametersTemp,Parameters,NumberOfParameters);

 f2=EvaluateCostFunction(NumberOfParameters, Parameters);
 InitialFunctionValue=f2;

 do {
    Accepted=0;

    *c = *c * HeatingFactor;

    for(Proposed=1;Proposed<=ProposedTransitions;Proposed++) {

      SetNewParameters(GenerateMode,
		       *c,
		       ParametersNew,
		       Parameters,
		       UpperBounds,
		       LowerBounds,
		       NumberOfParameters);

      f1=EvaluateCostFunction(NumberOfParameters,ParametersNew);

      /* new parameter set worse than old one */
      if(f2<f1) {
        if(Accept(ModeAccept,
		  *c,
		  f1, 
		  f2, 
		  FactorGeneralizedSA,
		  OptimalValueErrorFunction)) {
          ++Accepted;
	  f2=f1;
          CopyVector(Parameters, ParametersNew, NumberOfParameters); 
        }
      }
      else {
        ++Accepted;
	f2=f1;
        CopyVector(Parameters, ParametersNew, NumberOfParameters); 
      }
    }

    /* Prevent convergence -> restart cycly with same parameters */
    CopyVector(Parameters,ParametersTemp,NumberOfParameters);
    f2=InitialFunctionValue;

    fprintf(fp,"  c: %9.5lf Proposed: %d Accepted: %d AcceptanceRatio: %lf\n",
                *c,
                ProposedTransitions, 
	        Accepted,
	        AcceptanceRatio(ProposedTransitions,Accepted));
    fflush(fp);

 } while(AcceptanceRatio(ProposedTransitions,Accepted)<InitialAcceptanceRatio);
 
 /* restore initial parameters */
 CopyVector(Parameters,ParametersTemp,NumberOfParameters);

 FREE1D(ParametersTemp);
 FREE1D(ParametersNew);

}


/**************************************************************************/
/**                                                                      **/
/**                                                                      **/
/**                           ACCEPTANCE RATIO                           **/
/**                                                                      **/
/**                                                                      **/
/**************************************************************************/

/**************************************************************************/
/* Procedure: AcceptanceRatio()                                           */
/**************************************************************************/
double AcceptanceRatio(int Proposed,
		       int Accepted)
/*
 * Author: A.H.C. van Kampen, 14-7-93
 * Purpose:  calculates the acceptance ratio which is defined by         
 *           (Number of accepted transitions)/(Number of proposed transi)
 */
{
 return((double)Accepted/(double)Proposed);
}


/**************************************************************************/
/**                                                                      **/
/**                                                                      **/
/**                           ACCEPTANCE CRITERIA                        **/
/**                                                                      **/
/**                                                                      **/
/**************************************************************************/


/**************************************************************************/
/* Procedure: MetropolisAcceptanceCriterium                               */
/**************************************************************************/
int MetropolisAcceptanceCriterium(double f1,
				  double f2,
				  double c)
/*
 * Author: A.H.C. van Kampen, 15-7-93
 * Purpose: calculated the acceptance probability according to the  
 *          metropolis criterium
 */
{
 if(exp((f2-f1)/c)>RandomSA(&RandomFraction)) return(TRUE);
 return(FALSE);
}

/**************************************************************************/
/* Procedure: SigmoideAcceptanceCriterium()                               */
/**************************************************************************/
int SigmoideAcceptanceCriterium(double f1,
				double f2,
				double c)
/*
 * Author: A.H.C. van Kampen, 27-6-93
 * Purpose: calculated the acceptance probability by means of a sigmoide 
 *          function.
 *          Schuur, P.C., 'classification of acceptance criteria for the  
 *          simulated annealing algorithm', memorandum COSOR 89-29        
 *          vakgroep besliskunde en stochastiek, TU eindhoven, 1989  
 */
{
 if((1.0/(1.0+exp((f2-f1)/c)))>RandomSA(&RandomFraction)) return(TRUE);
 return(FALSE);
}

/**************************************************************************/
/* Procedure: GSAAcceptanceCriterium()                                    */
/**************************************************************************/
int GSAAcceptanceCriterium(double f1,
			   double f2,
			   double FactorGeneralizedSA,
			   double OptimalValueErrorFunction)
/*
 * Author: A.H.C. van Kampen, 27-6-93
 * Purpose:  Acceptance criterium for Generalized simulated annealing    
 *           This include only constant parameters, not c!  
 *           M.E. Johnson, "Generalized simulated annealing for function   
 *           optimization",Technometrics,28(3):209,1986  
 */
{
 if (exp(FactorGeneralizedSA*(f2-f1)/(f2-OptimalValueErrorFunction)) >
              RandomSA(&RandomFraction)) return(TRUE);
 return(FALSE);
}

/**************************************************************************/
/* Procedure: Accept()                                                    */
/**************************************************************************/
int Accept(int     Mode,
	   double  ControlParameter,
	   double  f1,
	   double  f2,
	   double  FactorGeneralizedSA,
	   double  OptimalValueErrorFunction)
/*
 * Author: A.H.C. van Kampen, 15-7-93
 * Purpose: determines which acceptance criterium to use    
 *          ModeString: string containing integer value >=0,<=9,    
 */
{
 switch(Mode) {
 case 1: 
   return(MetropolisAcceptanceCriterium(f1,f2,ControlParameter));
   break;

 case 2:
   return(SigmoideAcceptanceCriterium(f1, f2, ControlParameter));
   break;

 case 3:
   return(GSAAcceptanceCriterium(f1, 
				 f2, 
				 FactorGeneralizedSA, 
				 OptimalValueErrorFunction));
   break;
 }
}


/**************************************************************************/
/**                                                                      **/
/**                                                                      **/
/**                           EXIT MARKOV CHAIN                          **/
/**                                                                      **/
/**                                                                      **/
/**************************************************************************/


/**************************************************************************/
/* Procedure: ExitMarkovChain()                                           */
/**************************************************************************/
int ExitMarkovChain(char *ModeString)
/*
 * Author: A.H.C. van Kampen, 26-7-93
 * Purpose: determines which (compound) criterium is used to exit        
 *          markov chain. All exit criteria can be used simultaneously.
 *          In this case ModeString contrains more numbers. 
 *          For example to use to first two criteria: ModeString="12"      
 */
{
 if(
    (CheckMode(1,ModeString) &&
     ExitMarkovChainNumberOfTransitions(Accepted,
					MinimumNumberTransitions)) ||

    (CheckMode(2,ModeString) &&
     ExitMarkovChainMaximumTransitions(MarkovChainStep,
				       MaximumMarkovChain)) ||
    (CheckMode(3,ModeString) &&
     ExitMarkovChainFunctionValues(LastFunctionValues,
				   MarkovChainStep,
				   MaximumNumberConstantFunctionValues)) || 
    (CheckMode(4,ModeString) &&
     ExitMarkovChainMaximumRejections(Rejected,
				      MaximumRejections))

   ) return(TRUE);
 return(FALSE);
}

/**************************************************************************/
/* Procedure: ExitMarkovChainFunctionValue()                              */
/**************************************************************************/
int ExitMarkovChainFunctionValues(double *LastFunctionValues,
				  int     MarkovChainStep,
				  int     MaximumNumberConstantFunctionValues)
/*
 * Author: A.H.C. van Kampen, 26-7-93
 * Purpose: terminates the markov chain if the last n function values do 
 *          not differ         
 *          Kirkpartrick, S., J. Stat. Phys, 34, 975-986, 1984   
 */
{
 int i;

 for(i=0;i<MaximumNumberConstantFunctionValues-1;i++)
   if(LastFunctionValues[i]!=LastFunctionValues[i+1]) return(FALSE);

 fprintf(fp," \n>>Exit Markov chain: last %d function values\n",
	        MaximumNumberConstantFunctionValues);
 fprintf(fp," Did not change\n");

 /* reinitialize array */
 for(i=0;i<MaximumNumberConstantFunctionValues;i++)
   LastFunctionValues[i]=i;

 return(TRUE);
}

/**************************************************************************/
/* Procedure: ExitMarkovChainNumberOfTransistions()                       */
/**************************************************************************/
int ExitMarkovChainNumberOfTransitions(int Transition,
				       int MinimumNumberTransitions)
/*
 * Author: A.H.C. van Kampen, 15-7-93
 * Purpose:  exits the markov chain if the number of transitions is       
 *           higher than a minimum treshold 
 */
{
 if(Transition>=MinimumNumberTransitions) {
   fprintf(fp," \n>>Exit Markov chain: minimum number of transitions reached\n");
   fprintf(fp," Number of transitions %d\n",Transition);
   return(TRUE);
 }
 return(FALSE);
}

/**************************************************************************/
/* Procedure: ExitMarkovChainMaximumTransitions()                         */
/**************************************************************************/
int ExitMarkovChainMaximumTransitions(int MarkovChainStep,
				      int MaximumMarkovChain)
/*
 * Author: A.H.C. van Kampen, 15-7-93
 * Purpose: exits the markov chain if the number of transitions is      
 *          higher than a maximum treshold  
 *          this prevents too long markov chains when the control         
 *          parameter decreases and the ExitMarkovChainNumberOfTransitions
 *          takes too long to exit. 
 */
{
 if(MarkovChainStep>=MaximumMarkovChain) {
   fprintf(fp," \n>>Exit markov chain: chain to long\n");
   fprintf(fp," Length markov chain: %d\n",MarkovChainStep);
   fprintf(fp," Maximum value Markov chain: %d\n",MaximumMarkovChain);
   return(TRUE);
 }
 return(FALSE);
}

/**************************************************************************/
/* Procedure: ExitMarkovChainMaximumRejections()                          */
/**************************************************************************/
int ExitMarkovChainMaximumRejections(int Rejected,
				     int MaximumRejections)
/*
 * Author: A.H.C. van Kampen, 26-7-93
 * Purpose: exits the markov chain if the number of rejections is   
 *          higher than a minimum treshold   
 *          Simulated annealing: theory and applications'    
 */
{
 if(Rejected>=MaximumRejections) {
   fprintf(fp," \n>>Exit markov chain: Maximum number\n");
   fprintf(fp," of rejections exceeded.\n");
   fprintf(fp," Rejections: %d\n",Rejected);
   fprintf(fp," Maximum number of rejections: %d\n",MaximumRejections);
   return(TRUE);
 }
 return(FALSE);
}


/**************************************************************************/
/**                                                                      **/
/**                                                                      **/
/**                       EXIT SIMULATED ANNEALING                       **/
/**                                                                      **/
/**                                                                      **/
/**************************************************************************/


/**************************************************************************/
/* Procedure: ExitSimulatedAnnealing()                                    */
/**************************************************************************/
int ExitSimulatedAnnealing(char *ModeString)
/*
 * Author: A.H.C. van Kampen, 15-7-93
 * Purpose:  determines whether the SA is completed         
 *           ModeString: string containing integer values >=0, <=9     
 */
{
 if(
    (CheckMode(1,ModeString) &&
     ExitSimulatedAnnealingControlParameter(c, MinimumControlParameter)) ||

    (CheckMode(2,ModeString) &&
     ExitSimulatedAnnealingAcceptation(MarkovChainStep, /*Proposed*/
				       Accepted, 
				       MinimumAcceptanceRatio)) ||

    (CheckMode(3,ModeString) &&
     ExitSimulatedAnnealingFunctionValues(LastFunctionValues,
				MarkovChainStep,
				MaximumNumberConstantFunctionValues)) ||

    (CheckMode(4,ModeString) && 
     ExitSimulatedAnnealingMachinePrecision(c,f_new)) ||

    (CheckMode(5,ModeString) && 
     ExitSimulatedAnnealingMinimumCostFunction(f_new,
					       MinimumValueCostFunction)) 

   ) return(TRUE);
 return(FALSE);
}

/**************************************************************************/
/* Procedure: ExitSimulatedAnnealingMachinePrecision()                    */
/**************************************************************************/
int ExitSimulatedAnnealingMachinePrecision(double c,
					   double f1)
/*
 * Author: A.H.C. van Kampen, 16-7-93
 * Purpose: checks if the control parameter or the generated function  
 *          value is less than the MACHINE_PRECISION, if so the          
 *          simulated annealing is terminated.    
 */
{
 if((fabs(c)<=MACHINE_PRECISION) || (fabs(f1)<=MACHINE_PRECISION)) {
    fprintf(fp,"\n\nWarning: simulated annealing is exited because the\n");
    fprintf(fp,"calculated function value and/or control parameter\n");
    fprintf(fp,"is less than the MACHINE_PRECISION.\n\n");
    return(TRUE);
  }
 return(FALSE);
}

/**************************************************************************/
/* Procedure: ExitSimulatedAnnealingMinimumCostFunction()                 */
/**************************************************************************/
int ExitSimulatedAnnealingMinimumCostFunction(double f1,
					      double MinimumCostFunction)
/*
 * Author: A.H.C. van Kampen, 27-7-93
 * Purpose: checks if the function value is less than some minimum      
 *          treshold. If so, simulated annealing is terminated        
 */
{
 if(fabs(f1)<=MinimumCostFunction) {
    fprintf(fp,"\n\nWarning: simulated annealing is exited because the\n");
    fprintf(fp,"calculated function value %lf is less then the minimum\n");
    fprintf(fp,"value: %lf.\n\n",f1,MinimumCostFunction);
    return(TRUE);
  }
 return(FALSE);
}
 
/**************************************************************************/
/* Procedure: ExitSimulatedAnnealingAcception()                           */
/**************************************************************************/
int ExitSimulatedAnnealingAcceptation(int    Proposed,
				      int    Accepted,
				      double MinimumAcceptanceRatio)
/*
 * Author: A.H.C. van Kampen, 17-6-93
 * Purpose: terminated the SA if the acceptance ratio is less than a    
 *          certain minimum treshold        
 *          Barakat, M.T., Journal Comp-aided mol design, 4, 295-316 
 */
{
 if(AcceptanceRatio(Proposed,Accepted)<=MinimumAcceptanceRatio) {
   fprintf(fp," \n>>Exit simulated annealing: minimum acceptation ratio\n");
   fprintf(fp," is reached.\n");
   fprintf(fp," Minimum acceptation ratio: %lf\n", MinimumAcceptanceRatio);
   fprintf(fp," Acceptation ratio %lf\n",AcceptanceRatio(Proposed,Accepted));
   return(TRUE);
 }
 return(FALSE);
}

/**************************************************************************/
/* Procedure: ExitSimulatedAnnealingFunctionValues()                      */
/**************************************************************************/
int ExitSimulatedAnnealingFunctionValues(double *LastFunctionValues,
					 int     MarkovChainStep,
					 int     MaximumNumberConstantFunctionValues)
/*
 * Author: A.H.C. van Kampen, 26-7-93
 * Purpose: terminates SA if the last n function values do not differ
 *          Kirkpartrick, S., J. Stat. Phys, 34, 975-986, 1984  
 */
{
 int i;

 for(i=0;i<MaximumNumberConstantFunctionValues-1;i++)
   if(LastFunctionValues[i]!=LastFunctionValues[i+1]) return(FALSE);

 fprintf(fp," \n>>Exit simulated annealing: last %d function values\n",
	        MaximumNumberConstantFunctionValues);
 fprintf(fp," Did not change\n");
 return(TRUE);
}

/**************************************************************************/
/* Procedure: ExitSimulatedAnnealingControlParameter()                    */
/**************************************************************************/
int ExitSimulatedAnnealingControlParameter(double ControlParameter,
					   double MinimumControlParameter)
/*
 * Author: A.H.C. van Kampen, 15-7-93
 * Purpose:  exits simulated annealing when the control parameter      
 *           is less than a certain treshold     
 */
{
 if(ControlParameter<=MinimumControlParameter) {
   fprintf(fp," \n>>Exit simulated annealing: control parameter less than\n");
   fprintf(fp," Minimum treshold.\n");
   fprintf(fp," c: %lf\n",ControlParameter);
   fprintf(fp," Treshold %lf\n",MinimumControlParameter);
   return(TRUE);
 }
 return(FALSE);
}


/**************************************************************************/
/**                                                                      **/
/**                                                                      **/
/**                           GENERATING FUNCTIONS                       **/
/**                                                                      **/
/**                                                                      **/
/**************************************************************************/


/**************************************************************************/
/* Procedure: GenerateNewParametes()                                      */
/**************************************************************************/
void SetNewParameters(int     GenerateMode,
		      double  ControlParameter,
		      double *ParametersNew,
		      double *Parameters,
		      double *UpperBounds,
		      double *LowerBounds,
		      int     NumberOfParameters)

/*
 * Author: A.H.C. van Kampen, 26-7-93
 * Purpose:  generates new parameter according to a selected procedure     
 */
{
 int i;

   switch (GenerateMode) {
     case 1:
       GenerateNewParametersGauss(ParametersNew,
				  Parameters,
				  UpperBounds,
				  LowerBounds,
				  NumberOfParameters,
				  ControlParameter);
       break;
     
     case 2:
       GenerateNewParametersCauchy(ParametersNew,
				   Parameters,
				   UpperBounds,
				   LowerBounds,
				   NumberOfParameters,
				   ControlParameter);
       break;

     case 3:
       GenerateNewParametersUniform(ParametersNew,
				    Parameters,
				    UpperBounds,
				    LowerBounds,
				    NumberOfParameters,
				    ControlParameter);
       break;

     case 4: 
       GenerateNewParametersUser(ParametersNew,
				 Parameters,
				 UpperBounds,
				 LowerBounds,
				 NumberOfParameters,
				 ControlParameter);
       break;
   }

}

/**************************************************************************/
/* Procedure: GenerateNewParametersUniform()                              */
/**************************************************************************/
void GenerateNewParametersUniform(double *ParametersNew,
				  double *Parameters,
				  double *UpperBounds,
				  double *LowerBounds,
				  int     NumberOfParameters,
				  double  ControlParameter)
/*
 * Author: A.H.C. van Kampen, 26-7-93
 * Purpose: generates new parameters according to p=p +/- U(0,1)) 
 *          where U(0,1) is a uniform distribution between 0 and 1
 *          these function tries to generate a new parameter value. If    
 *          it does not succeed in 1000 trials, it terminates and the     
 *          new parameters will be equal to the old ones.   
 */

{
 int i, Trial;
 double x1,Sum, *Direction, RF;
 boole  flag;

 CALLOC1D(Direction,NumberOfParameters);

 Trial=0;
 do {
   ++Trial;
   
   Sum=0.0;
   for(i=0;i<NumberOfParameters;i++) {
     x1=RANDOM_BOUND_DOUBLE(1.0,&RandomFraction);
     Direction[i]=x1;
     Sum += x1*x1;
     x1=RANDOM_BOUND_DOUBLE(1.0,&RF);
     if(x1>Direction[i]) Direction[i]=-Direction[i];
   }

   for(i=0;i<NumberOfParameters;i++) 
     ParametersNew[i]= Parameters[i] + StepSizeParameters * Direction[i]/sqrt(Sum);

   flag=FALSE;
   for(i=0;i<NumberOfParameters;i++) 
     if(ParametersNew[i]>UpperBounds[i] || ParametersNew[i]<LowerBounds[i]) {
       ParametersNew[i]=Parameters[i];        
       flag=TRUE;
     }

 } while(flag && Trial!=1000); 
 
 if(Trial==1000) {
   printf("\nParameter generation: no new parameters are generated\n");
   printf("because this didn't succeeded in 1000 trials\n\n");
 }

 FREE1D(Direction);
} 

/**************************************************************************/
/* Procedure: GenerateNewParametersCauchy()                               */
/**************************************************************************/
void GenerateNewParametersCauchy(double *ParametersNew, 
				 double *Parameters,
				 double *UpperBounds,
				 double *LowerBounds,
				 int     NumberOfParameters,
				 double  ControlParameter)
/*
 * Author: A.H.C. van Kampen, 28-7-93
 * Purpose:  generates new parameters. The deviation from the old        
 *           parameters is calculated according to a cauchy distribution      
 *           these function tries to generate a new parameter value. If   
 *           it does not succeed in 1000 trials, it terminates and the     
 *           new parameters will be equal to the old ones.                 
 *           Ingber, L., 'Genetic algorithms and very fast simulated       
 *                                                                        
 *           Here an x1 is generated with mean 0 and a deviation 1        
 *                                                                        
 *           Szu, H. 'Fast simulated annealing', Physics Letters A,        
 *           122(3,4), 1987.                                               
 *                                                                        
 *           Ingber, L., 'Genetic algorithms and very fast simulated       
 *           reannealing: a comparison', Mathl. Comput. Modelling,16(11)   
 *           87-100,1992  
 */
{
 int i, Trial;
 double x1, Sum, *Direction;
 boole  flag;

 CALLOC1D(Direction,NumberOfParameters);

 Trial=0;
 do {
   ++Trial;
   
   Sum=0.0;
   for(i=0;i<NumberOfParameters;i++) {
     x1=tan(PI*(RandomSA(&RandomFraction)-0.5));
     Direction[i]=x1;
     Sum += x1*x1;
   }

   for(i=0;i<NumberOfParameters;i++) 
     ParametersNew[i]= Parameters[i] + StepSizeParameters * Direction[i]/sqrt(Sum);

   flag=FALSE;
   for(i=0;i<NumberOfParameters;i++) 
     if(ParametersNew[i]>UpperBounds[i] || ParametersNew[i]<LowerBounds[i]) {
       ParametersNew[i]=Parameters[i];        
       flag=TRUE;
     }
 } while(flag && Trial!=1000); 
 
 if(Trial==1000) {
   printf("\nParameter generation: no new parameters are generated\n");
   printf("because this didn't succeeded in 1000 trials\n\n");
 }

 FREE1D(Direction);
} 

/**************************************************************************/
/* Procedure: GenerateNewParametersGauss()                                */
/**************************************************************************/
void GenerateNewParametersGauss (double *ParametersNew, 
				 double *Parameters,
				 double *UpperBounds,
				 double *LowerBounds,
				 int     NumberOfParameters,
				 double  ControlParameter)

/*
 * Author: A.H.C. van Kampen, 30-7-93
 * Purpose:  generates new parameters. The deviation from the old         
 *           parameters is calculated according to a Gauss distribution  
 *           with a N(0,1).   
 *           This function tries to generate a new parameter value. If
 *           it does not succeed in 1000 trials, it terminates and the    
 *           new parameters will be equal to the old ones.                
 *                                                                       
 *           The function StandardGauss0() generated two numbers which are
 *           distributed with mean 0 and standard deviation 1.0.
 *                                                                       
 */

{
 int i, Trial;
 double x1,x2, Sum, *Direction;
 boole  flag;

 CALLOC1D(Direction,NumberOfParameters);

 Trial=0;
 do {
   ++Trial;
   
   Sum=0.0;
   for(i=0;i<NumberOfParameters;i++) {
     SA_StandardGauss0(&x1,&x2,&RandomFraction);
     Direction[i]=x1;
     Sum += x1*x1;
   }

   for(i=0;i<NumberOfParameters;i++) 
     ParametersNew[i]= Parameters[i] + StepSizeParameters * Direction[i]/sqrt(Sum);
 
   flag=FALSE;
   for(i=0;i<NumberOfParameters;i++) 
     if(ParametersNew[i]>UpperBounds[i] || ParametersNew[i]<LowerBounds[i]) {
       ParametersNew[i]=Parameters[i];        
       flag=TRUE;
     }

 } while(flag && Trial!=1000); 
 
 if(Trial==1000) {
   printf("\nParameter generation: no new parameters are generated\n");
   printf("because this didn't succeeded in 1000 trials\n\n");
 }

 FREE1D(Direction);
} 



    
